package com.wjholden.nmap;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.ProgressDialog;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.pm.PackageManager.NameNotFoundException;
import android.net.Uri;
import android.net.wifi.WifiInfo;
import android.net.wifi.WifiManager;
import android.os.AsyncTask;
import android.os.Bundle;
import android.os.Handler;
import android.os.Looper;
import android.os.Message;
import android.preference.PreferenceManager;
import android.text.format.Formatter;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.AdapterView;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.EditText;
import android.widget.Spinner;
import android.widget.TextView;

/**
 * <a href="http://nmap.wjholden.com">Nmap for Android!</a>
 * The original versions used Nmap binaries I got from Robert McCurdy (http://rmccurdy.com/stuff/G1/BINS/NMAP/).
 * Around v0.3 or Vlatko Kosturjak build newer binaries based on Nmap 5.3 (http://ftp.linux.hr/android/nmap/).
 * @author William John Holden (wjholden@gmail.com)
 * @version 0.4.7
 */
public class NmapMain extends Activity implements NmapConstants {
	/**
	 * Text area where results are shown
	 */
	private transient TextView mResults;
	
	/**
	 * Input area where command-line arguments (other than target IP)
	 */
	private transient EditText mArguments;
	
	/**
	 * Input area where user supplies scan target
	 */
	private transient EditText mTarget;
	
	/**
	 * Start button initiates scan 
	 */
	private transient Button mStart;
	
	/**
	 * Help button, shows result of [command] --help
	 */
	private transient Button mHelp;
	
	/**
	 * Share button, launches Compose Email activity with message body of any data in mResults
	 */
	private transient Button mShare;
	
	/**
	 * Command spinner switches between Nmap, Nping, and Ncat
	 */
	private transient Spinner mCommandSpinner;
	
	/**
	 * Switches between Nmap, XML, and Grepable Nmap output
	 */
	private transient Spinner mOutputSpinner;

	/**
	 * if "VerifyInstallation()" method has run this should be set to true
	 * @since 0.4.2
	 * This and most other static variables here are saved in SharedPreferences so verifyInstallation
	 * only runs once.
	 */
	public static boolean installVerified;
	
	/**
	 * tells us whether one or more scans have been executed
	 */
	public static boolean hasRunOneScan;
	
	/**
	 * tells us if "su" command is available
	 */
	public static boolean canGetRoot;
	
	/**
	 * this is where the nmap/nping/ncat compiled binaries should be stored.
	 */
	public static String bindir;
	
	/**
	 * location of the 'su' command, if present, otherwise 'sh'.
	 */
	public static String sh;
	
	/**
	 * the command to be executed (nmap, nping, ncat) and optional (and mandatory) command line arguments from mArguments
	 */
	public static String outputArgs;

	/**
	 * tells the scanning engine what command is to be executed (nmap, nping, ncat)
	 */
	public static int scanType;
	
	/**
	 * verify installation thread
	 */
	private AsyncTask<Object, Void, Void> vTask = null;
	
	/**
	 * scanning engine thread
	 */
	private AsyncTask<Object,Void,Void> sTask = null;
	
	/**
	 * handler for inter-thread communication
	 */
	public static Handler handler;
	
	/**
	 * Static so that the other classes can turn it off after it runs.
	 * This shall be the only ProgressDialog used. Don't use "new" - instead just change the text
	 * and show/hide it.
	 */
	public static ProgressDialog pd;
	
	/**
	 * Store both user and application settings in here.
	 */
	private SharedPreferences settings;
	
	/**
	 * Use this to save application and user settings into the SharedPreferences.
	 */
	private SharedPreferences.Editor preferencesEditor;
	
	/**
	 * This variable is set by the user with the context menu.
	 * If on, use the Log utility to send debugging information to
	 * logcat, and also present user with a menu option to e-mail
	 * debugging output to developer.
	 */
	public static boolean debug;
	
	/**
	 * Instance of my custom error logging utility, designed to store
	 * (instead of just stream) debugging output for troubleshooting.
	 */
	public static NmapError ne;
	
	/**
	 * Only used get get IP address and SSID.
	 */
	private WifiManager wifiManager;
	
	/**
	 * Provides methods for getting IP address and SSID.
	 */
	private WifiInfo wifiInfo;

	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		setContentView(R.layout.main);
		
		ne = new NmapError(); // this kind of needs to happen early on, ha.

		initScreen();
		
		settings = PreferenceManager.getDefaultSharedPreferences(this.getApplicationContext());
		String target = settings.getString("target", ""); // originally ("target", null)
		/** @since 22 - just set target = "" if not in settings...hints should work for "" */
		//if (target != null && !target.equals("")) // these checks are to enable hints for new users 
		mTarget.setText(target);
		String args = settings.getString("args", ""); // originally ("args", null)
		/** @since 22 same as above */
		//if (args != null && !args.equals(""))
		mArguments.setText(args);
		/** @since 22 moved license notice to res/values/strings.xml */
		final String resultText = settings.getString("result", getString(R.string.gpl));
		mResults.setText(resultText);
		
		installVerified = settings.getBoolean("installVerified", false);
		ne.log("installVerified = " + installVerified);
		hasRunOneScan = settings.getBoolean("hasRunOneScan", false);
		ne.log("hasRunOneScan = " + hasRunOneScan);
		canGetRoot = settings.getBoolean("canGetRoot", NmapUtilities.checkRootPermissions());
		ne.log("canGetRoot = " + canGetRoot);
		sh = settings.getString("sh", null);
		ne.log("sh = " + sh);
		
		/** @since 22 bindir is no longer a setting and is generated at runtime every execution for safety
		 * This may end up being an expensive call but I think it's worth it for reliability */
		//bindir = settings.getString("bindir", null);
		bindir = NmapUtilities.findBinDir(NmapUtilities.getDataDirectory(getApplicationContext()));
		final File myBindir = new File(bindir);
		if (!myBindir.exists() && !myBindir.mkdirs())
		{
			final Message msg = Message.obtain();
			msg.obj = "Critical Error: unable to create " + bindir;
			msg.what = INSTALL_ERROR;
			NmapMain.handler.sendMessage(msg);
		}
		
		/** @since 21 - do NOT put outputArgs into SharedPreference! It broke the whole thing earlier. */
		/** @since 22 - use the SD Card instead of the ../tmp/ folder...this should make it good */
		/** @since 22 - move this into the scan class as it needs to be different for different programs */
		//Nmap.outputArgs = new String(" -oA " + "/sdcard/" + mCommandSpinner.getSelectedItem().toString() + " ");
		
		debug = settings.getBoolean("DEBUG", false);
		
		try {
			int curVersion = getPackageManager().getPackageInfo("com.wjholden.nmap", 0).versionCode;
			NmapMain.ne.log("Nmap version: " + curVersion);
			NmapMain.ne.log("SDK version: " + android.os.Build.VERSION.SDK);
		} catch (NameNotFoundException e) {
			NmapMain.ne.log(e.toString());
		}
		
		initHandler();

		if (!installVerified && (vTask == null || vTask.getStatus() == AsyncTask.Status.FINISHED || vTask.getStatus() == AsyncTask.Status.RUNNING))
		{
			pd = new ProgressDialog(this);
			pd.setMessage("Verifying installation. Select \"Allow\" if prompted.");
			pd.show();
			vTask = new NmapInstall().execute(this.getApplicationContext(), "norestart");
		}

		showWifiInfo();
		
		mStart.setOnClickListener(new View.OnClickListener() {

			@Override
			public void onClick(View v) {
				//scanType = mCommandSpinner.getSelectedItemPosition() + SCANTYPE_NMAP;

				mResults.setText("");
				
				final String targetAndArguments = mTarget.getText().toString() + " " + mArguments.getText().toString();
				/*if (targetAndArguments == null || targetAndArguments.length() == 0)
					targetAndArguments = "";*/
				if (sTask == null || sTask.getStatus() == AsyncTask.Status.FINISHED || sTask.getStatus() == AsyncTask.Status.RUNNING)
				{
					mStart.setText("Abort");
					pd = new ProgressDialog(NmapMain.this);
					pd.setMessage("Please wait...");
					pd.show();
					sTask = new NmapScan().execute(getApplicationContext(), targetAndArguments);
					if (NmapMain.debug)
					{
						NmapMain.ne.log("Starting scan with the following values:");
						NmapMain.ne.log("\t" + targetAndArguments);
					}
				}
				else
				{
					if (!sTask.cancel(false))
					{
						handler.sendEmptyMessage(THREAD_ERROR);
					}
				}
			}
		});

		mHelp.setOnClickListener(new View.OnClickListener() {

			@Override
			public void onClick(View v) {
				//scanType = mCommandSpinner.getSelectedItemPosition() + SCANTYPE_NMAP;
				pd = new ProgressDialog(NmapMain.this);
				pd.setMessage("Press your menu button to get even more documentation.");
				pd.show();
				
				mResults.setText("");
				if (sTask == null || sTask.getStatus() == AsyncTask.Status.FINISHED || sTask.getStatus() == AsyncTask.Status.RUNNING)
				{
					sTask = new NmapScan().execute(getApplicationContext(), "-h");
				}
				else
				{
					if (!sTask.cancel(false))
					{
						handler.sendEmptyMessage(THREAD_ERROR);
					}
				}
			}
		});

		mShare.setOnClickListener(new View.OnClickListener() {

			@Override
			public void onClick(View v) {
				final Intent emailIntent = new Intent(android.content.Intent.ACTION_SEND);
				emailIntent.setType("plain/text");
				emailIntent.putExtra(android.content.Intent.EXTRA_SUBJECT, "Nmap Scan Results");
				emailIntent.putExtra(android.content.Intent.EXTRA_TEXT, mResults.getText().toString());
				NmapMain.this.startActivity(emailIntent);
			}
		});
	}

	public static void setBinDir(final String b)
	{
		bindir = new String(b);
	}
	
	public static void setSh(final String s)
	{
		sh = new String(s);
	}
	
	@Override
	protected void onDestroy() {
		if (vTask != null)
		{
			vTask.cancel(true);
		}
		if (sTask != null)
		{
			sTask.cancel(true);
		}
		saveUserEntry();
		super.onDestroy();
	}

	protected void onResume(Bundle savedInstanceState) {
		super.onRestoreInstanceState(savedInstanceState);
		installVerified = savedInstanceState.getBoolean("installVerified");
		hasRunOneScan = savedInstanceState.getBoolean("hasRunOneScan");
		canGetRoot = savedInstanceState.getBoolean("canGetRoot");
		bindir = savedInstanceState.getString("bindir");
		outputArgs = savedInstanceState.getString("outputArgs");
		sh = savedInstanceState.getString("sh");
		if (vTask != null)
			vTask.cancel(true);
		if (sTask != null)
			sTask.cancel(true);
	}

	protected void onPause(Bundle outState) {
		saveUserEntry();
		
		Looper.myLooper().quit();
		outState.putBoolean("installVerified", installVerified);
		outState.putBoolean("hasRunOneScan", hasRunOneScan);
		outState.putBoolean("canGetRoot", canGetRoot);
		outState.putString("bindir", bindir);
		outState.putString("outputArgs", outputArgs);
		outState.putString("sh", sh);
		super.onSaveInstanceState(outState);
	}
	
	private void saveUserEntry()
	{
		preferencesEditor = settings.edit();
		preferencesEditor.putString("target", mTarget.getText().toString());
		preferencesEditor.putString("args", mArguments.getText().toString());
		preferencesEditor.putString("result", mResults.getText().toString());
		preferencesEditor.putBoolean("DEBUG", debug);
		
		preferencesEditor.putBoolean("installVerified", installVerified);
		preferencesEditor.putBoolean("hasRunOneScan", hasRunOneScan);
		preferencesEditor.putBoolean("canGetRoot", canGetRoot);
		preferencesEditor.putString("bindir", bindir);
		// remove outputArgs @since 21
		// preferencesEditor.putString("outputArgs", outputArgs);
		preferencesEditor.putString("sh", sh);
		
		preferencesEditor.commit();
	}

	@Override
	public boolean onPrepareOptionsMenu(Menu menu) {
		menu.removeGroup(MENU_GROUP_DEBUG);
		menu.removeGroup(MENU_GROUP_EXTRA_OPTIONS);
		if (NmapMain.debug)
		{
			menu.add(MENU_GROUP_DEBUG, MENU_ITEM_DEBUG_ENABLE, Menu.NONE, "Disable Debugging");
			menu.add(MENU_GROUP_DEBUG, MENU_ITEM_DEBUG_EMAIL, Menu.NONE, "Email debugging output to developer");
		}
		else
		{
			menu.add(MENU_GROUP_DEBUG, MENU_ITEM_DEBUG_ENABLE, Menu.NONE, "Enable Debugging");
		}
		
		menu.add(MENU_GROUP_DEBUG, MENU_ITEM_DATALOG, Menu.NONE, "Survey");
		menu.add(MENU_GROUP_EXTRA_OPTIONS, MENU_ITEM_MORE_HELP, Menu.NONE, "More help on the web");
		menu.add(MENU_GROUP_EXTRA_OPTIONS, MENU_ITEM_REINSTALL, Menu.NONE, "Reinstall binaries");
		menu.add(MENU_GROUP_EXTRA_OPTIONS, MENU_ITEM_EXIT, Menu.NONE, "Exit");
		
		return super.onCreateOptionsMenu(menu);
	}

	@Override
	public boolean onMenuItemSelected(int featureId, MenuItem item) {
		switch (item.getItemId())
		{
		case MENU_ITEM_DEBUG_ENABLE:
			NmapMain.debug ^= true;
			break;
		case MENU_ITEM_DEBUG_EMAIL:
			final Intent emailIntent = new Intent(android.content.Intent.ACTION_SEND);
			emailIntent.setType("plain/text");
			emailIntent.putExtra(android.content.Intent.EXTRA_SUBJECT, "Nmap Debugging Output");
			emailIntent.putExtra(android.content.Intent.EXTRA_TEXT, NmapMain.ne.getLog());
			emailIntent.putExtra(android.content.Intent.EXTRA_EMAIL, new String[] { "wjholden@gmail.com" });
			NmapMain.this.startActivity(emailIntent);
			break;
		case MENU_ITEM_MORE_HELP:
			Uri uri = Uri.parse( "http://nmap.org/book/man.html" );
			startActivity( new Intent( Intent.ACTION_VIEW, uri ) );
			break;
		case MENU_ITEM_EXIT:
			this.finish();
			break;
		case MENU_ITEM_REINSTALL:
			pd = new ProgressDialog(this);
			pd.setMessage("Verifying installation. Select \"Allow\" if prompted.");
			pd.show();
			vTask = new NmapInstall().execute(this.getApplicationContext(), "restart");
			break;
		case MENU_ITEM_DATALOG:
			NmapMain.ne.log("Showing 'Survey' Activity.");
			Intent surveyActivity = new Intent(NmapMain.this, NmapSurvey.class);
			startActivity(surveyActivity);
			break;
		default:
			NmapMain.ne.log("Somehow got the 'default' option in onMenuItemSelected.");
			break;
		}
		return super.onMenuItemSelected(featureId, item);
	}
	
	private void initScreen()
	{
		mResults = (TextView) findViewById(R.id.results);
		mArguments = (EditText) findViewById(R.id.Arguments);
		mTarget = (EditText) findViewById(R.id.Target);
		mStart = (Button) findViewById(R.id.Start);
		mHelp = (Button) findViewById(R.id.Help);
		mShare = (Button) findViewById(R.id.Share);
		
		mCommandSpinner = (Spinner) findViewById(R.id.CommandSpinner);
		ArrayAdapter<CharSequence> adapterCommand = ArrayAdapter.createFromResource(
				this, R.array.commands, android.R.layout.simple_spinner_item);
		adapterCommand.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
		mCommandSpinner.setAdapter(adapterCommand);

		mCommandSpinner.setOnItemSelectedListener(new AdapterView.OnItemSelectedListener() {

			@Override
			public void onItemSelected(AdapterView<?> arg0, View arg1,
					int arg2, long arg3) {
				scanType = mCommandSpinner.getSelectedItemPosition() + SCANTYPE_NMAP;

				/*if (mCommandSpinner.getSelectedItem().toString().equals("nmap") && !mArguments.getText().toString().contains("--system-dns"))
					mArguments.setText("--system-dns " + mArguments.getText().toString()); */

				NmapMain.ne.log(mCommandSpinner.getSelectedItem().toString() + " selected.");
			}

			@Override
			public void onNothingSelected(AdapterView<?> arg0) {
				NmapMain.ne.log("Nothing was selected in mCommandSpinner.");
			}});
		
		mOutputSpinner = (Spinner) findViewById(R.id.OutputSpinner);
		ArrayAdapter<CharSequence> adapterOutput = ArrayAdapter.createFromResource(
				this, R.array.output, android.R.layout.simple_spinner_item);
		adapterOutput.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
		mOutputSpinner.setAdapter(adapterOutput);
		
		mOutputSpinner.setOnItemSelectedListener(new AdapterView.OnItemSelectedListener() {

			@Override
			public void onItemSelected(AdapterView<?> arg0, View arg1,
					int arg2, long arg3) {
				if (hasRunOneScan && installVerified)
				{					
					mResults.setText("");
					handler.sendEmptyMessage(RUN_COMPLETE);
				}
			}

			@Override
			public void onNothingSelected(AdapterView<?> arg0) {
				NmapMain.ne.log("Nothing was selected in mOutputSpinner.");
			}
		});
	}
	
	/**
	 *  Try to get the SSID and IP. If you get it, set the users title to reflect this information.
	 */
	private void showWifiInfo()
	{
		wifiManager = (WifiManager) getSystemService(WIFI_SERVICE);
		wifiInfo = wifiManager.getConnectionInfo();
		String ssid = wifiInfo.getSSID();
		String ip = Formatter.formatIpAddress(wifiInfo.getIpAddress());
		if (ssid != null && ip != null)
			this.setTitle(new String(this.getString(R.string.app_name) + "  SSID: " + ssid + "  IP: " + ip));
	}
	
	/**
	 * Added 0.4.7. Forces the activity to restart. Used after user has manually reinstalled binaries.
	 */
	private void forceRestart()
	{
		Intent intent = this.getIntent();
		this.finish();
		startActivity(intent);
	}
	
	private void initHandler()
	{
		handler = new Handler() {
			@Override
			public void handleMessage(Message msg)
			{
				AlertDialog.Builder alert = new AlertDialog.Builder(NmapMain.this).setPositiveButton("OK", new DialogInterface.OnClickListener() {
					
					@Override
					public void onClick(DialogInterface dialog, int which) {
						dialog.dismiss();
					}
				});;
				
				if (msg.obj == null)
				{
					NmapMain.ne.log("(h-code: " + msg.what + ")");
				}
				else
				{
					NmapMain.ne.log((String) msg.obj + " (h-code: " + msg.what + ")");
				}
				
				switch (msg.what)
				{
				case INSTALL_NO_ROOT:
				case RUN_ERROR:
					NmapMain.ne.log("Runtime Error");
				case INSTALL_ERROR:
					mResults.setText((String) msg.obj);
					alert.setMessage((String) msg.obj);
					alert.show();
					mStart.setText("Start");
					break;
				case INSTALL_GOOD:
					alert.setMessage("This is not an official release by www.nmap.org.\nPlease request support only from wjholden@gmail.com.");
					alert.show();
					break;
				case RUN_LINE:
					mResults.setText(mResults.getText() + "\n" + (String) msg.obj);
					break;
				case RUN_COMPLETE:
					try {
						String type = mOutputSpinner.getSelectedItem().toString();
						if (type.equals("Normal"))
						{
							type = new String("nmap");
						}
						else if (type.equals("XML"))
						{
							type = new String("xml");
						}
						else if (type.equals("Grepable"))
						{
							type = new String("gnmap");
						}
						else
						{
							NmapMain.ne.log("Something went wrong with mOutputSpinner and received type: " + type);
							type = new String("nmap");
						}
						BufferedReader b = new BufferedReader(
								new FileReader(
										NmapUtilities.verifyTemporaryFolder(getApplicationContext()) + 
										"/" +
										mCommandSpinner.getSelectedItem().toString() + "." +
										type
										)
								);
						String line;
						while ((line = b.readLine()) != null)
						{
							mResults.setText(mResults.getText() + line + "\n");
							NmapMain.ne.log(line);
						}
						b.close();
					} catch (FileNotFoundException e) {
						// @since 22 Because users keep getting weird errors when it gets to here I've decided to show the license.
						// We'll see how this works.
						//mResults.setText(getString(R.string.gpl));
						Message msg1 = Message.obtain();
						msg1.obj = e.toString();
						msg1.what = RUN_ERROR;
						handler.sendMessage(msg1);
					} catch (IOException e) {
						Message msg1 = Message.obtain();
						msg1.obj = e.toString();
						msg1.what = RUN_ERROR;
						handler.sendMessage(msg1);
					} catch (NameNotFoundException e) {
						Message msg1 = Message.obtain();
						msg1.obj = e.toString();
						msg1.what = RUN_ERROR;
						handler.sendMessage(msg1);
					}
					// removed 0.3.6 this was annoying once I added the nmap/xml/gnmap selection
					//alert.setMessage("Scan Complete!");
					//alert.show();
					mStart.setText("Start");
					break;
				case THREAD_ERROR:
					alert.setMessage("Unable to cancel task.");
					alert.show();
					break;
				case FORCE_RESTART:
					NmapMain.ne.log("Forcing Activity restart...");
					forceRestart();
					break;
				default:
					ne.log("Unexpected case in handler.");
					break;
				}
			}
		};
	}
}
